/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
#include "license.readme"

#include "nsCRT.h"
#include "nsFileStream.h"
#include "xp_core.h"
#include "myspAffixmgr.h"
#include "myspAffix.h"
#include "nsIFile.h"
#include "nsReadCLine.h"
#include "nsReadableUtils.h"
#include "nsDirectoryServiceDefs.h"


/* Modifications for mozilla Copyright 2001 David Einstein Deinst@world.std.com  */


myspAffixMgr::myspAffixMgr() 
{
  pTable = new myspPrefix*[MAXAFFIXES];
  cappfx =(pTable?MAXAFFIXES:0);

  sTable = new myspSuffix*[MAXAFFIXES];
  capsfx =(sTable?MAXAFFIXES:0);

  numpfx = 0;
  numsfx = 0;
}


myspAffixMgr::~myspAffixMgr() 
{
  mPersonalDictionary = nsnull;
  for (PRUint32 i=0;i<numpfx;i++)
    delete (pTable[i]);
  delete [] pTable;
  
  for (PRUint32 j=0;j<numsfx;j++)
    delete (sTable[j]);
  delete [] sTable;

  numpfx = 0;
  numsfx = 0;
}

nsresult myspAffixMgr::GetPersonalDictionary(mozIPersonalDictionary * *aPersonalDictionary)
{
  *aPersonalDictionary = mPersonalDictionary;
  NS_IF_ADDREF(*aPersonalDictionary);
  return NS_OK;
}

nsresult myspAffixMgr::SetPersonalDictionary(mozIPersonalDictionary * aPersonalDictionary)
{
  mPersonalDictionary = aPersonalDictionary;
  return NS_OK;
}

nsresult 
myspAffixMgr::Load(const nsString& aDictionary)
{
  nsresult res=NS_OK;
  nsCOMPtr<nsIFile> dicFile;
  nsCOMPtr<nsIFile> affFile;
  PRBool fileExists;

  //get the directory
  res = NS_GetSpecialDirectory(NS_XPCOM_COMPONENT_DIR, getter_AddRefs(dicFile));
  if(NS_FAILED(res)) return res;
  if(!dicFile)return NS_ERROR_FAILURE;
  res = dicFile->Append("myspell");
  if(NS_FAILED(res)) return res;
  res = dicFile->Exists(&fileExists);
  if(NS_FAILED(res)) return res;
  if(!fileExists) return NS_ERROR_FAILURE; 
  res = dicFile->Clone(getter_AddRefs(affFile));
  if(NS_FAILED(res)) return res;
  if(!dicFile)return NS_ERROR_FAILURE;

  //get the affix file
  nsAutoString affName;
  affName=aDictionary;
  affName.Append(NS_LITERAL_STRING(".aff"));
  res=affFile->AppendUnicode(affName.get());
  if(NS_FAILED(res)) return res; 
  res = affFile->Exists(&fileExists);
  if(NS_FAILED(res)) return res;
  if(!fileExists) return NS_ERROR_FAILURE; 

  //get the dictionary file
  nsAutoString dicName;
  dicName=aDictionary;
  dicName.Append(NS_LITERAL_STRING(".dic"));
  res=dicFile->AppendUnicode(dicName.get());
  if(NS_FAILED(res)) return res; 
  res = dicFile->Exists(&fileExists);
  if(NS_FAILED(res)) return res;
  if(!fileExists) return NS_ERROR_FAILURE; 
 
  // load the affixFile
  nsCOMPtr<nsIInputStream> affStream;
  res = NS_NewLocalFileInputStream(getter_AddRefs(affStream), affFile);
  if(NS_FAILED(res)) return res;
  if(!affStream)return NS_ERROR_FAILURE;
  res = parse_file(affStream);


  res = mPersonalDictionary->SetCharset(mEncoding.get());
  if(NS_FAILED(res)) return res;

  PRInt32 pos=aDictionary.FindChar('-');
  if(pos<1) pos = 2;  // FIXME should be min of 2 and aDictionary.Length()
  nsAutoString lang;
  lang.Assign(Substring(aDictionary,0,pos));
  res = mPersonalDictionary->SetLanguage(lang.get());
  if(NS_FAILED(res)) return res;


  // load the dictionary
  nsCOMPtr<nsIInputStream> dicStream;
  res = NS_NewLocalFileInputStream(getter_AddRefs(dicStream), dicFile);
  if(NS_FAILED(res)) return res;
  if(!dicStream)return NS_ERROR_FAILURE;
  res = LoadDictionary(dicStream);

  return res;
}


// read in aff file and build up prefix and suffix data structures
nsresult  myspAffixMgr::parse_file(nsIInputStream *strm)
{

  PRInt32 i, j;
  PRInt32 numents;
  nsLineBuffer *lineBuffer;
  nsresult res;
  res= NS_InitLineBuffer(&lineBuffer);
  nsCAutoString line;
  PRBool moreData=PR_TRUE;
  PRInt32 pos;
  nsACString::const_iterator begin,end;
  PRInt32 opos=0;
  // step one is to parse the affix file building up the internal
  // affix data structures

  numpfx = 0;       // number of prefixes
  numsfx = 0;       // number of suffixes
  numents = 0;      // number of affentry structures to parse
  char flag='\0';   // affix char identifier
  {  
    PRInt16 ff=0;
    char ft;


    // read in each line ignoring any that do not
    // start with PFX or SFX

    while (moreData) {
      NS_ReadLine(strm,lineBuffer,line,&moreData);
      /* parse in the try string */
      if (Substring(line,0,3).Equals("TRY")) {
        pos = line.FindChar(' ');
        if(pos != -1){
          trystring = Substring(line,pos+1,line.Length()-pos-1);
        } 
      }

      /* parse in the name of the character set used by the .dict and .aff */
      if (Substring(line,0,3).Equals("SET")) {

        pos = line.FindChar(' ');
        if(pos != -1){
          nsCAutoString cencoding;
          cencoding.Assign(Substring(line,pos+1,line.Length()-pos-1));
          mEncoding.AssignWithConversion(cencoding.get());
        } 
      }

      // get the type of this affix: P - prefix, S - suffix
      ft = ' ';
      if (Substring(line,0,3).Equals("PFX")) ft = 'P';
      if (Substring(line,0,3).Equals("SFX")) ft = 'S';
      if (ft != ' ') {
        i = 0;

        // split line into pieces

        do {
          pos=line.FindChar(' ',PR_FALSE,opos);
          if (pos != opos) {
            switch(i) 
              {
                // piece 1 - is type of affix
              case 0: break;
          
                // piece 2 - is affix char
              case 1:  
                line.BeginReading(begin);
                begin.advance(opos);
                flag = *begin; 
                break; 

                // piece 3 - is cross product indicator 
              case 2: 
                line.BeginReading(begin);
                begin.advance(opos);
                if (*begin == 'Y') 
                  ff = XPRODUCT; 
                break; 

                // piece 4 - is number of affentries
              case 3: 
                line.BeginReading(begin);
                begin.advance(opos);
                if(pos == -1){
                  line.EndReading(end);
                }
                else {
                  line.BeginReading(end);
                  end.advance(pos);
                }
                numents = atoi(PromiseFlatCString(Substring(begin,end)).get()); 
                break;

              default: 
                break;
              }
            i++;
          }
          opos=pos+1;
        }while(pos != -1);


        myspAffix *affptr=nsnull;
        // now create the correct Affix Object (Prefix or Suffix) FIXME do not fix the affixes.
        if (ft == 'P') {
          if(numpfx < cappfx){
            affptr = pTable[numpfx++] =  new myspPrefix(this,flag,numents,ff);
          }
          else {
            myspPrefix **temp=new myspPrefix*[2*cappfx+1];
            if(temp){
              PRUint32 i;
              for(i=0;i<numpfx;i++) temp[i]=pTable[i];
              delete [] pTable;
              pTable = temp;
              cappfx = 2*cappfx+1;
              affptr = pTable[numpfx++] =  new myspPrefix(this,flag,numents,ff);
            }
          }
        } else {
          if(numsfx < capsfx){
            affptr = sTable[numsfx++] = new myspSuffix(this,flag,numents,ff);
          }
          else {
            myspSuffix **temp=new myspSuffix*[2*capsfx+1];
            if(temp){
              PRUint32 i;
              for(i=0;i<numsfx;i++) temp[i]=sTable[i];
              delete [] sTable;
              sTable = temp;
              capsfx = 2*capsfx+1;
              affptr = sTable[numsfx++] =  new myspSuffix(this,flag,numents,ff);
            }
          }
        }
        if(affptr){
          // now parse numents affentries for this affix
          for (j=0; (j < numents)&&moreData; j++) {
            NS_ReadLine(strm,lineBuffer,line,&moreData);
            affptr->SetAffentry(j,line);
          }
        }
         
        numents = 0;
      }
    }
  }
  return NS_OK;
}


nsresult
myspAffixMgr::LoadDictionary(nsIInputStream *strm)
{
  nsLineBuffer *lineBuffer;
  nsresult res;
  res= NS_InitLineBuffer(&lineBuffer);
  nsCAutoString line;
  PRBool moreData;
  PRInt32 pos;
  nsACString::const_iterator begin1,end1,begin2,end2;
  
  // first read the first line of file to get hash table size */

  res = NS_ReadLine(strm,lineBuffer,line,&moreData);
  
  // loop through all words on much list and add to hash
  // table and create word and affix strings

  while (moreData) {
    res = NS_ReadLine(strm,lineBuffer,line,&moreData);
    
    // split each line into word and affix char strings
    pos = line.FindChar('/');
    if(pos==-1){
      line.BeginReading(begin1);
      line.EndReading(end1);
      begin2=end2=begin1;
    }
    else{
      line.BeginReading(begin1);
      begin2=end1=begin1;
      end1.advance(pos);
      begin2.advance(pos+1);
      line.EndReading(end2);
    }


    // add the word and its index
    mHashTable.Put(PromiseFlatCString(Substring(begin1,end1)).get(),PromiseFlatCString(Substring(begin2,end2)).get());

  }

  return NS_OK;
}






/* cross check suffixes with a specific prefix */
const char * myspAffixMgr::cross_check (const nsAFlatCString &word, 
                                        int sfxopts, myspPrefix * ppfx)
{
  PRUint32 i=0;
  const char * rv= NULL;
  // go through every possible suffix with this prefix
  while (i < numsfx)  {
    rv = sTable[i]->check(word, sfxopts, ppfx); 
    if (rv) return rv;
    i++;
  }
  return NULL;
}



// check if word with affixes is correctly spelled
const char * myspAffixMgr::affix_check (const nsAFlatCString &word)
{
  PRUint32 i;
  const char * rv= NULL;
    
  // first check all prefixes (also cross check with suffixes if need be)
  i=0;
  while (i < numpfx) {
    rv = pTable[i]->check(word); 
    if (rv) return rv;
    i++;
  }

  // if still not found check all suffixes
  i = 0;
  while (i < numsfx) {
    rv = sTable[i]->check(word, 0, NULL); 
    if (rv) return rv;
    i++;
  }

  return NULL;
}


// return text encoding of dictionary
nsString myspAffixMgr::get_encoding()
{
  return mEncoding;
}


// return the preferred try string for suggestions
nsCString myspAffixMgr::get_try_string()
{
  return trystring;
}

PRBool myspAffixMgr::check(const nsAFlatCString &word)
{
  const char * he = NULL;

  he = lookup(word);

  if ((he == NULL)) {
    // try stripping off affixes */
    he = affix_check(word);
  }

  if (he) return PR_TRUE;
  PRBool good=PR_FALSE;
  nsresult res = mPersonalDictionary->Check(word.get(),&good);
  if(NS_FAILED(res))
    return PR_FALSE;
  return good;
}


// utility method to look up root words in hash table
const char * myspAffixMgr::lookup(const nsAFlatCString &word)
{
  return mHashTable.Get(word.get());
}



